use detee_shared::general_proto::brain_general_cli_server::BrainGeneralCliServer;
use detee_shared::vm_proto::brain_vm_cli_server::BrainVmCliServer;
use surreal_brain::grpc::BrainGeneralCliMock;
use surreal_brain::grpc::BrainVmCliMock;
use surreal_brain::db;
use tonic::transport::{Identity, Server, ServerTlsConfig};

#[tokio::main]
async fn main() {
    env_logger::builder().filter_level(log::LevelFilter::Debug).init();
    db::init().await.unwrap();
    let addr = "0.0.0.0:31337".parse().unwrap();

    let snp_cli_server = BrainVmCliServer::new(BrainVmCliMock {});
    let general_service_server = BrainGeneralCliServer::new(BrainGeneralCliMock {});

    let cert = std::fs::read_to_string("./tmp/brain-crt.pem").unwrap();
    let key = std::fs::read_to_string("./tmp/brain-key.pem").unwrap();

    let identity = Identity::from_pem(cert, key);

    Server::builder()
        .tls_config(ServerTlsConfig::new().identity(identity))
        .unwrap()
        .add_service(snp_cli_server)
        .add_service(general_service_server)
        .serve(addr)
        .await
        .unwrap();
}
